/**
 * Copyright (c) 2021 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#pragma once
#include <stdint.h>
#include <malloc.h>
#include "../../engine/display/primitive.h"
#include "../../engine/display/texture.h"
#include "../../engine/display/tileset.h"

////////////////////////////////////////////////////////////////////////////////
// Entities
////////////////////////////////////////////////////////////////////////////////

/** Entity ID */
typedef uint8_t entityid_t;

typedef struct {
  float x, y, z;
} entitypos_t;


////////////////////////////////////////////////////////////////////////////////
// Tiles
////////////////////////////////////////////////////////////////////////////////

/** Width of a tile (in pixels) */
#define TILE_WIDTH 16
/** Height of a tile (in pixels) */
#define TILE_HEIGHT 16

/** Bitwise Flags from tiles. */
typedef uint8_t tileflag_t;

/** Tile ID */
typedef uint8_t tileid_t;

/** Representation of a unique tile within a chunk. */
typedef struct {
  /** ID of the tile */
  tileid_t id;

  /** Rendering indice and vertice offsets for the tile. */
  int32_t indiceStart, verticeStart;
} tile_t;

/** Representation of the information of a tile within a tilemap. */
typedef struct {
  /** Flags of the tile */
  tileflag_t flags;

  /** How many indices and vertices a tile with this definition has. */
  int32_t indiceCount, verticeCount;
} tiledef_t;

/** Representation of the tile definitions within a tilemap. */
typedef struct {
  /** Tile definitions within the tilemap */
  tiledef_t *tileDefinitions;
} tilemap_t;


////////////////////////////////////////////////////////////////////////////////
// Chunks
////////////////////////////////////////////////////////////////////////////////

/** Width (in tiles) of chunks. */
#define CHUNK_WIDTH 16
/** Heihgt (in tiles) of chunks. */
#define CHUNK_HEIGHT 16
/** Depth (in tiles) of chunks. */
#define CHUNK_DEPTH 8
/** Count of tiles in the chunk. */
#define CHUNK_TILE_COUNT CHUNK_WIDTH*CHUNK_HEIGHT*CHUNK_DEPTH



/** Representation of a chunk, a group of tiles that can be buffered around. */
typedef struct chunk_t {
  /** Position (in absolute chunk coordinates) of this chunk */
  int32_t x, y, z;

  /** Array of tiles within the chunk */
  tile_t tiles[CHUNK_TILE_COUNT];

  /** Ready to be rendered chunk 3d primitive */
  primitive_t *primitive;
} chunk_t;


////////////////////////////////////////////////////////////////////////////////
// Worlds
////////////////////////////////////////////////////////////////////////////////
/** Width of world (in chunks) */
#define WORLD_WIDTH 5
/** Height of world (in chunks) */
#define WORLD_HEIGHT WORLD_WIDTH
/** Depth of world (in chunks) */
#define WORLD_DEPTH 2

/** Count of chunks in the world */
#define WORLD_CHUNK_COUNT WORLD_WIDTH*WORLD_HEIGHT*WORLD_DEPTH

/** Representation of the world. */
typedef struct {
  /** Asset subdir name */
  char *assetWorldDirectory;

  /** Tileset texture */
  texture_t *texture;
  /** Tileset predivided */
  tileset_t *tileset;
  /** Tilemap of the world */
  tilemap_t *tilemap;

  /** Current (chunk) coordinates of the first chunk in the chunk list */
  int32_t x, y, z;
  /** Current chunk list, ordered */
  chunk_t *chunkList[WORLD_CHUNK_COUNT];
  /** Chunk array (unordered) */
  chunk_t chunks[WORLD_CHUNK_COUNT];
} world_t;